#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""General purpose math utility functions.

"""

import math
from math import ceil, floor, exp, log, sqrt
from math import degrees, radians
from math import sin, cos, tan
from math import asin, acos, atan, atan2

import numpy as np

__all__ = [
    # TRIG FUNCTIONS
    'sin', 'cos', 'tan', 'degrees', 'radians',

    # INVERSE TRIG FUNCTIONS
    'asin', 'acos', 'atan', 'atan2',

    # TRIG CONSTANTS
    'TWO_PI', 'PI', 'HALF_PI', 'QUARTER_PI', 'TAU', 'HALF_TAU',

    # MATH FUNCTIONS FROM THE STANDARD LIBRARY (abs and round are
    # available in builtins.)
    'ceil', 'floor', 'exp', 'log', 'sqrt',

    # MATH FUNCTIONS DEFINED HERE

    'constrain', 'lerp', 'remap', 'normalize', 'distance', 'dist',
    'magnitude', 'mag', 'sq',
]

TWO_PI = 2 * math.pi
PI = math.pi
HALF_PI = math.pi / 2.0
QUARTER_PI = math.pi / 4.0

TAU = 2 * math.pi
HALF_TAU = math.pi

# We will be using these a lot, just precompute a whole lot of sin and
# cosine values.
SINCOS_PRECISION = 0.5
SINCOS_LENGTH = int(360 / SINCOS_PRECISION)

PRE_SIN = [ sin(radians(d) * SINCOS_PRECISION) for d in range(SINCOS_LENGTH) ]
PRE_COS = [ cos(radians(d) * SINCOS_PRECISION) for d in range(SINCOS_LENGTH) ]

SINCOS = list(zip(PRE_SIN, PRE_COS))


def _sanitize(point, target_dimension=3):
    return list(point) + [0] * (target_dimension - len(point))

def _is_numeric(val):
    return isinstance(val, int) or isinstance(val, float)

def constrain(amount, low, high):
    """Constrain the given value in the specified range.

    Examples ::

        >>> constrain(8, 1, 5)
        5

        >>> constrain(5, 1, 5)
        5

        >>> constrain(3, 1, 5)
        3

        >>> constrain(1, 1, 5)
        1

        >>> constrain(-3, 1, 5)
        1

    :param amount: The the value to be contrained.

    :param low: The lower constain.

    :param high: The upper constain.

    """
    if amount < low:
        return low
    elif amount > high:
        return high
    else:
        return amount

def lerp(start, stop, amount):
    """Linearly interpolate the start value to the stop value.

    Examples ::

        >>> lerp(0, 10, 0.0)
        0.0

        >>> lerp(0, 10, 0.5)
        5.0

        >>> lerp(0, 10, 0.8)
        8.0

        >>> lerp(0, 10, 1.0)
        10.0

    :param start: The start value

    :param stop: The stop value

    :param amount: The amount by which to interpolate. (:math:`0 \leq
        amount \leq 1`).
    :type amount: float

    """
    return start + amount * (stop - start)

def remap(value, source_range, target_range):
    """Remap a value from the source range to the target range.

    Examples ::

         >>> remap(50, (0, 100), (0, 10))
         5.0

         >>> remap(5, (0, 10), (0, 100))
         50.0

         >>> remap(5, (0, 10), (10, 20))
         15.0

         >>> remap(15, (10, 20), (0, 10))
         5.0

    :param value: The value to be remapped.

    :param source_range: The source range for :code:`value`
    :type source_range: tuple

    :param target_range: The target range for :code:`value`
    :type target_range: tuple

    """
    s0, s1 = source_range
    t0, t1 = target_range
    S = s1 - s0
    T = t1 - t0
    return t0 + ((value - s0) / S) * T

def normalize(value, low, high):
    """Normalize the given value to the specified range.

    Examples ::

        >>> normalize(10, 0, 100)
        0.1

        >>> normalize(0.3, 0, 1)
        0.3

        >>> normalize(100, 0, 100)
        1.0

        >>> normalize(1, 1, 15)
        0.0

    :param value:
    :type value: float

    :param low: The lower bound for the range.
    :type low: float

    :param high: The upper bound for the range.
    :type high: float
    """
    return remap(value, (low, high), (0, 1))

def magnitude(x, y, z=0):
    """Return the magnitude of the given vector.

    Examples ::

        >>> magnitude(3, 4)
        5.0

        >>> magnitude(2, 3, 6)
        7.0

        >>> magnitude(0, 0, 0)
        0.0

    :param x: The x-component of the vector.
    :type x: float

    :param y: The y-component of the vector.
    :type y: float

    :param z: The z-component of the vector (defaults to 0).
    :type z: float

    :returns: The magnitude of the vector.
    :rtype: float

    """
    return np.sqrt(np.sum(np.array([x, y, z]) ** 2))

def distance(point_1, point_2):
    """Return the distance between two points.

    Examples ::

        >>> distance((0, 0, 0), (2, 3, 6))
        7.0

        >>> distance((2, 3, 6), (2, 3, 6))
        0.0

        >>> distance((6, 6, 6), (2, 3, 6))
        5.0

    :param point_1:
    :type point_1: tuple

    :param point_2:
    :type point_2: tuple

    :returns: The distance between two points
    :rtype: float

    """
    p1 = np.array(_sanitize(point_1))
    p2 = np.array(_sanitize(point_2))
    return np.sqrt(np.sum((p1 - p2) ** 2))

def sq(number):
    """Square a number.

    Examples ::

        >>> square(-25)
        625

        >>> square(0)
        0

        >>> square(13)
        169


    :param number: The number to be squared.
    :type number: float

    :returns: The square of the number.
    :rtype: float

    """
    return number ** 2


# Helpful aliases
dist = distance
mag = magnitude



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Shaders used by the main program"""

from collections import namedtuple

ShaderSource = namedtuple('ShaderSource', 'vert frag')

# vertex shader
default_vertex_source = """
attribute vec3 position;
attribute vec4 color;

varying vec4 frag_color;

uniform mat4 projection;
uniform mat4 perspective_matrix;

void main()
{
    gl_Position = projection * perspective_matrix * vec4(position, 1.0);
    frag_color = color;
}
"""

default_fragment_source = """
varying vec4 frag_color;

void main()
{
    gl_FragColor = frag_color;
}
"""


# Shader sources to draw framebuffers textues.
fbuffer_vertex_source = """
attribute vec2 position;
attribute vec2 texcoord;

varying vec2 vert_tex_coord;

void main() {
    gl_Position = vec4(position, 0, 1);
    vert_tex_coord = texcoord;
}
"""

fbuffer_fragment_source = """
uniform sampler2D texture;
varying vec2 vert_tex_coord;

void main() {
    gl_FragColor = texture2D(texture, vert_tex_coord.st);
}
"""

src_default = ShaderSource(default_vertex_source, default_fragment_source)
src_fbuffer = ShaderSource(fbuffer_vertex_source, fbuffer_fragment_source)



import unittest
import numpy as np

import xml.etree.ElementTree as etree
from p5.core.svg import (get_style, parse_rect, parse_circle, parse_line, parse_ellipse, parser)
from p5.core.color import Color
from p5.core.shape import PShape

svg = '<?xml version="1.0" encoding="utf-8"?> <!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20010904//EN" "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd"> \
    <svg width="400px" height="300px" xml:lang="fr" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"> \
        <line x1="100" x2="20" y1="20" y2="200" style="fill:none;stroke:springgreen;stroke-width:2px;"/> \
        <ellipse cx="200" cy="150" rx="50" ry="130" style="fill:none;stroke:lightsteelblue;stroke-width:30px;stroke-opacity:0.5;"/> \
        <circle cx="200" cy="150" r="50"/> \
        <rect x="0" y="0" width="300" height="100"/> \
    </svg>'

tree = etree.fromstring(svg)
elements = []
for e in tree:
    elements.append(e)

line = elements[0]
rect = elements[3]

# not tested because they use Renderer() class while creation of PShape()
ellipse = elements[1]
circle = elements[2]

class TestSVG(unittest.TestCase):
    def test_get_style(self):
        self.assertEqual(get_style(line, "stroke"), 'springgreen')
        self.assertEqual(get_style(line, "stroke-width"), 2)

        self.assertEqual(get_style(ellipse, "stroke"), "lightsteelblue")
        self.assertEqual(get_style(ellipse, "stroke-width"), 30)
        self.assertEqual(get_style(ellipse, "stroke-opacity"), 0.5)

    def test_parse_rect(self):
        parsed = parse_rect(rect)
        self.assertTrue(np.array_equal(parsed.vertices, np.array([(0, 0), (300, 0), (300, 100), (0, 100)])))
        self.assertTrue(np.array_equal(parsed._fill, Color(0.0, 0.0, 0.0, 0.0)))
        self.assertTrue(np.array_equal(parsed.stroke_weight, 1))
        self.assertTrue(np.array_equal(parsed.stroke,  Color(0.0, 0.0, 0.0, 0.0)))
        self.assertTrue(np.array_equal(parsed.stroke_join, 0))

    def test_parse_line(self):
        parsed = parse_line(line)

        self.assertTrue(np.array_equal(parsed.vertices, np.array([(100, 20), (20, 200)])))
        self.assertTrue(np.array_equal(parsed._fill, Color("none")))
        self.assertTrue(np.array_equal(parsed.stroke_weight, 2))
        self.assertTrue(np.array_equal(parsed.stroke,  Color("springgreen")))

if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from .utils import *
from .vector import *
from .rand import *
from .curves import *
from .time import *
from . import matrix



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import math

import numpy as np

def _magnitude(arr):
    """Return the madnitude of the given array.

    :param arr: Input array.
    :type arr: np.ndarray

    :returns: The magnitude of the input array.
    :rtype: float
    """
    return np.sqrt(arr.dot(arr))

def _normalize(arry):
    """Return the normalized version of the given array.

    :param arry: Input array to be normalized.
    :type arry: np.ndarray

    :returns: Normalized version of the input array.
    :rtype: np.ndarray

    """
    mag = _magnitude(arry)
    return arry / mag

def scale_transform(x, y, z=1):
    """Return a scale transformation matrix.

    :param x: Scale factor in the x direction.
    :type x: numeric

    :param y: Scale factor in the y direction.
    :type y: numeric

    :param z: Scale factor in the z direction.
    :type z: numeric (defaults to 1)

    :returns: A scale transformation matrix.
    :rtype: np.ndarray

    """
    scale_matrix = np.identity(4)
    scale_matrix[0, 0] = x
    scale_matrix[1, 1] = y
    scale_matrix[2, 2] = z
    return scale_matrix

def translation_matrix(x, y, z=0):
   """Return a new translation matrix.

   :param x: translation in the x-direction.
   :type x: numeric

   :param y: translation in the y-direction.
   :type y: numeric

   :param z: translation in the z-direction.
   :type z: numeric (defaults to 0)

   :returns: A transform matrix with the given translation applied to
       it
   :rtype: np.ndarray

   """
   translate_matrix = np.identity(4)
   translate_matrix[0, -1] = x
   translate_matrix[1, -1] = y
   translate_matrix[2, -1] = z
   return translate_matrix

def rotation_matrix(axis, angle):
    """Return a roation matrix with the given angle and rotation.

    :param axis: The axis along which the matrix should be rotated.
    :type axis: np.ndarray

    :param angle: Angle by which to rotate (in radians).
    :type angle: float

    :returns: A rotation matrix along the given axis and angle.
    :rtype: np.ndarray

    """
    x, y, z = _normalize(axis)
    s = np.sin(angle)
    c = np.cos(angle)
    c1 = 1 - c

    rotation = np.identity(4)

    rotation[0, 0] = x * x * c1 + c
    rotation[0, 1] = x * y * c1 - z * s
    rotation[0, 2] = x * z * c1 + y * s
    rotation[1, 0] = y * x * c1 + z * s
    rotation[1, 1] = y * y * c1 + c
    rotation[1, 2] = y * z * c1 - x * s
    rotation[2, 0] = x * z * c1 - y * s
    rotation[2, 1] = y * z * c1 + x * s
    rotation[2, 2] = z * z * c1 + c

    return rotation

def euler_rotation_matrix(heading, attitude, bank):
    """

    :returns: A rotation matrix based on the given parameters.
    :rtype: np.ndarray
    """
    pass

def triple_axis_rotation_matrix(x, y, z):
    """Return a rotation matrix based on three axes.

    :param x: x-axis for the rotation matrix.
    :type x: np.ndarray

    :param y: y-axis for the rotation matrix.
    :type y: np.ndarray

    :param z: z-axis for the rotation matrix.
    :type z: np.ndarray

    :returns: A rotation matrix based on the given parameters.
    :rtype: np.ndarray
    """
    rotation_matrix = np.identity(4)
    rotation_matrix[0, :3] = x
    rotation_matrix[1, :3] = y
    rotation_matrix[2, :3] = z
    return rotation_matrix

def look_at(eye, at, up):
    """Return a new 'look-at' matrix.

    :param eye: Location of the 'eye'
    :type eye: np.ndarray

    :param at: The point being looked at.
    :type at: np.ndarray

    :param up: Vector specifying the up-direction.
    :type up: np.ndarray

    :returns: A new look at matrix.
    :rtype: np.ndarray
    """
    z = _normalize(eye - at)
    x = _normalize(np.cross(up, z))
    y = np.cross(z, x)

    mat = triple_axis_rotation_matrix(x, y, z)
    mat.transpose()

    mat[0, 3] = (-1) * x.dot(eye)
    mat[1, 3] = (-1) * y.dot(eye)
    mat[2, 3] = (-1) * z.dot(eye)

    return mat

def perspective_matrix(field_of_view, aspect_ratio, near_plane, far_plane):
    """Return a perspective matrix.

    :param field_of_view: The field_of_view to use.
    :type field_of_view: float

    :param aspect_ratio: The aspect ratio of the projection.
    :type aspect_ratio: float

    :param near_plane: Location of the near plane of the projection.
    :type near_plane: float

    :param far_plane: Location of the far plane of the projection.
    :type far_plane: float

    :returns: A new perspective matrix.
    :rtype: np.ndarray

    :raises AssertionError: When the near plane is zero or the near
         plane is the same as the far plane.

    """
    assert not math.isclose(near_plane, 0), \
        "Near plane cannot be at zero."
    assert not math.isclose(near_plane, far_plane), \
        "Near plane and far plane can't be same"
    f = 1 / math.tan(field_of_view / 2)
    mat = np.identity(4)
    mat[0, 0] = f / aspect_ratio
    mat[1, 1] =  f
    mat[2, 2] = (far_plane + near_plane) / (near_plane - far_plane)
    mat[2, 3] = 2 * far_plane * near_plane / (near_plane - far_plane)
    mat[3, 2] = -1
    mat[3, 3] = 0
    return mat



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import textwrap

from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont

from .image import image
from .image import PImage
from .structure import push_style

from . import p5

__all__ = ['create_font', 'load_font', 'text', 'text_font',
	'text_align', 'text_leading', 'text_size', 'text_width',
	'text_ascent', 'text_descent'
	]

_font_family = ImageFont.load_default()
_text_align_x = "LEFT"
_text_align_y = "TOP"
_text_leading = 0

def create_font(name, size=10):
	"""Create the given font at the appropriate size.
	
	:param name: Filename of the font file (only pil, otf and ttf 
		fonts are supported.)
	:type name: str

	:param size: Font size (only required when `name` refers to a
		truetype font; defaults to None)
	:type size: int | None

	"""

	if name.endswith('ttf') or name.endswith('otf'):
		font = ImageFont.truetype(name, size)
	elif name.endswith('pil'):
		font = ImageFont.load(name)
	else:
		raise NotImplementedError("Font type not supported.")

	return font

def load_font(font_name):
	"""Loads the given font into a font object

	"""
	return create_font(font_name)

def text(text_string, position, wrap_at=None):
	"""Draw the given text on the screen and save the image.

	:param text_string: text to display
	:type text_string: str

	:param position: position of the text on the screen
	:type position: tuple

	:param wrap_at: specifies the text wrapping column (defaults to
		None)
	:type wrap_at: int

	:returns: actual text that was drawn to the image (when wrapping
		is not set, this is just the unmodified text_string)
	:rtype: str

	"""

	if len(text_string) == 0:
		return 

	global _font_family, _text_leading
	
	multiline = False
	if not (wrap_at is None):
		text_string = textwrap.fill(text_string, wrap_at)
		size = _font_family.getsize_multiline(text_string)
		multiline = True
	elif "\n" in text_string:
		multiline = True
		size = list(_font_family.getsize_multiline(text_string))
		size[1] += _text_leading*text_string.count("\n")
	else:
		size = _font_family.getsize(text_string)
		
	canvas = Image.new("RGBA", size, color=(0, 0, 0, 0))
	canvas_draw = ImageDraw.Draw(canvas)
	
	if multiline:
		canvas_draw.multiline_text((0, 0), text_string, font=_font_family, spacing=_text_leading)
	else:
		canvas_draw.text((0, 0), text_string, font=_font_family)

	text_image = PImage(*size)
	text_image._img = canvas

	width, height = size
	position = list(position)
	if _text_align_x == "LEFT":
		position[0] += 0
	elif _text_align_x == "RIGHT":
		position[0] -= width
	elif _text_align_x == "CENTER":
		position[0] -= width/2

	if _text_align_y == "TOP":
		position[1] += 0
	elif _text_align_y == "BOTTOM":
		position[1] -= height
	elif _text_align_y == "CENTER":
		position[1] -= height/2

	with push_style():
		if p5.renderer.fill_enabled:
			p5.renderer.tint_enabled = True
			p5.renderer.tint_color = p5.renderer.fill_color
		image(text_image, position)
	
	return text_string

def text_font(font, size=10):
	"""Set current text font.

	:param font:
	:type font: PIL.ImageFont.ImageFont

	"""
	global _font_family
	_font_family = font

def text_align(align_x, align_y=None):
	"""Set the alignment of drawing text

	:param align_x: "RIGHT", "CENTER" or "LEFT".
	:type align_x: string

	:param align_y: "TOP", "CENTER" or "BOTTOM".
	:type align_y: string

	"""

	global _text_align_x, _text_align_y
	_text_align_x = align_x

	if align_y:
		_text_align_y = align_y

def text_leading(leading):
	"""Sets the spacing between lines of text in units of pixels

	:param leading: the size in pixels for spacing between lines
	:type align_x: int

	"""

	global _text_leading
	_text_leading = leading

def text_size(size):
	"""Sets the current font size

	:param leading: the size of the letters in units of pixels
	:type align_x: int

	"""

	global _font_family

	# reload the font with new size
	if hasattr(_font_family, 'path'):
		if _font_family.path.endswith('ttf') or _font_family.path.endswith('otf'):
			_font_family = ImageFont.truetype(_font_family.path, size)
	else:
		raise ValueError("text_size is nor supported for Bitmap Fonts")

def text_width(text):
	"""Calculates and returns the width of any character or text string

	:param text_string: text
	:type text_string: str

	:returns: width of any character or text string
	:rtype: int

	"""

	return _font_family.getsize(text)[0]

def text_ascent():
	"""Returns ascent of the current font at its current size

	:returns: ascent of the current font at its current size
	:rtype: float

	"""
	global _font_family
	ascent, descent = _font_family.getmetrics()
	return ascent

def text_descent():
	"""Returns descent of the current font at its current size

	:returns:  descent of the current font at its current size
	:rtype: float

	"""

	global _font_family
	ascent, descent = _font_family.getmetrics()
	return descent



import unittest
from p5.core.color import Color

class TestColor(unittest.TestCase):
    def test_parser(self):
        # RGB
        self.assertEqual(Color(100, 150, 200).red, 100)
        self.assertEqual(Color(100, 150, 200).green, 150)
        self.assertEqual(Color(100, 150, 200).blue, 200)
        self.assertEqual(Color(100, 150, 200).alpha, 255)

        # Greyscale
        self.assertEqual(Color(255), Color(255, 255, 255, 255))
        self.assertEqual(Color(0), Color(0, 0, 0, 255))

        # Grey Alpha
        self.assertEqual(Color(255, 100), Color(255, 255, 255, 100))
        self.assertEqual(Color(0, 100), Color(0, 0, 0, 100))

        # HEX
        self.assertEqual(Color("#fcba03"), Color(252, 186, 3))

        # Name
        self.assertEqual(Color("crimson"), Color(220, 20, 60))
        
if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Shaders used by the main program"""

from collections import namedtuple

ShaderSource = namedtuple('ShaderSource', 'vert frag')

# vertex shader
default_vertex_source = """
attribute vec3 position;
attribute vec4 color;

varying vec4 frag_color;

uniform mat4 modelview;
uniform mat4 projection;

void main()
{
    gl_Position = projection * modelview * vec4(position, 1.0);
    frag_color = color;
}
"""

default_fragment_source = """
varying vec4 frag_color;

void main()
{
    gl_FragColor = frag_color;
}
"""

# texture vertex shader
texture_vertex_source = """
attribute vec2 position;
attribute vec2 texcoord;

uniform mat4 transform;
uniform mat4 modelview;
uniform mat4 projection;

varying vec4 vertex_texcoord;

void main()
{
    gl_Position = projection * modelview * transform * vec4(position, 0.0, 1.0);
    vertex_texcoord = vec4(texcoord, 1.0, 1.0);
}
"""

# texture fragment shader
texture_fragment_source = """
uniform vec4 fill_color;
uniform sampler2D texture;

varying vec4 vertex_texcoord;

void main()
{
    gl_FragColor = texture2D(texture, vertex_texcoord.st) * fill_color;
}
"""

# Shader sources to draw framebuffers textues.
fbuffer_vertex_source = """
attribute vec2 position;
attribute vec2 texcoord;

varying vec2 vert_tex_coord;

void main() {
    gl_Position = vec4(position, 0, 1);
    vert_tex_coord = texcoord;
}
"""

fbuffer_fragment_source = """
uniform sampler2D texture;
varying vec2 vert_tex_coord;

void main() {
    gl_FragColor = texture2D(texture, vert_tex_coord.st);
}
"""

# vertex shader
stroke_vertex_source = """
attribute vec3 pos;
attribute vec3 posPrev;
attribute vec3 posCurr;
attribute vec3 posNext;
attribute vec4 color;

attribute float marker;
attribute float linewidth;
attribute float side;
attribute float join_type;
attribute float cap_type;

uniform mat4 modelview;
uniform mat4 projection;

varying vec4 frag_color;
varying float v_linewidth;
varying float v_join_type;
varying float v_cap_type;
varying float v_length;
varying float v_join; // > 1 if the vertex is stroke join
varying float v_cap;  // > 1 if the vertex is stroke cap 

varying vec3 v_tangentNext;
varying vec3 v_tangentPrev;

varying vec3 v_pos;

void main()
{   

    float width = 1.0;

    bool join = false;
    bool cap = false;

    float pi = 3.1415926535897;

    if(linewidth > 1){ // if width of line < 1, render line of width 1
        width = linewidth;
    }

    vec3 tangentPrev = posPrev - posCurr;
    vec3 tangentNext = posNext - posCurr;
    vec3 lineTangent;
    vec3 outsideTangent;

    if(side > 0.0){ // is this left/right side of line segment
        lineTangent = normalize(tangentNext);
        outsideTangent = normalize(tangentPrev);
    } else{
        lineTangent = normalize(tangentPrev);
        outsideTangent = normalize(tangentNext);
    }

    // calculate a vector perpendicular to the line
    vec3 perpendicular = normalize(vec3(lineTangent.y, -lineTangent.x, 0.0));

    float angle = acos(dot(lineTangent, outsideTangent));
    vec3 bisector = normalize(lineTangent + outsideTangent);
    float alignment = dot(perpendicular, outsideTangent);

    float factor = 1; // multiplication factor to scale the length of bisector

    if(side > 0.0){ // left side
        if(alignment > 0.0){
            if(marker > 0.0){
                perpendicular = bisector;
                factor = cos(pi/2 - angle/2);
            } else{
                perpendicular = bisector;
                factor = cos(pi/2 - angle/2);
            }
        } else{
            if(marker > 0.0){
                perpendicular = -bisector;
                factor = cos(pi/2 - angle/2);
            } else{
                perpendicular = -bisector;
                factor = cos(pi/2 - angle/2);
            }
        }
    } else{ // right side
        if(alignment > 0.0){
            if(marker > 0.0){
                perpendicular = bisector;
                factor = cos(pi/2 - angle/2);
            } else{
                perpendicular = bisector;
                factor = cos(pi/2 - angle/2);
            }
        } else{
            if(marker > 0.0){
                perpendicular = -bisector;
                factor = cos(pi/2 - angle/2);
            } else{
                perpendicular = -bisector;
                factor = cos(pi/2 - angle/2);
            }
        }
    }

    // For cap or straight edge
    vec3 offset = vec3(0.0, 0.0, 0.0);
    if(tangentPrev.x == 0 && tangentPrev.y == 0){
        factor = 1.0;
        v_cap = 1.0;
        perpendicular = normalize(vec3(lineTangent.y, -lineTangent.x, 0.0));
        offset = -width/2*lineTangent;
    } else if(tangentNext.x == 0 && tangentNext.y == 0){
        factor = 1.0;
        v_cap = 1.0;
        perpendicular = normalize(vec3(lineTangent.y, -lineTangent.x, 0.0));
        offset = -width/2*lineTangent;
    } else if(length(bisector) == 0.0){
        v_cap = -1.0;
        factor = 1.0;
        perpendicular = normalize(vec3(lineTangent.y, -lineTangent.x, 0.0));
    } else{
        v_cap = -1.0;
    }

    gl_Position = projection * modelview * vec4(posCurr + offset + marker*width*perpendicular/2/factor, 1.0);
    frag_color = color;

    //Set vertex shader variables
    v_linewidth = width;
    v_cap_type = cap_type;
    v_join_type = join_type;
    v_pos = pos;

    if(side > 0.0){ // left side
        v_tangentNext = posNext - posCurr;
        v_tangentPrev = posPrev - posCurr;
        v_length = length(v_tangentNext);
    } else{ // right side
        v_tangentNext = posCurr - posPrev;
        v_tangentPrev = posCurr - posNext;
        v_length = length(v_tangentNext);
    }    
}
"""

stroke_fragment_source = """
varying vec4 frag_color;
varying float v_linewidth;
varying float v_cap; // render cap or join

varying float v_join_type;
varying float v_cap_type;
varying vec3 v_pos;

varying vec3 v_tangentNext; // tangents for stroke join and cap methods
varying vec3 v_tangentPrev; // tangents for stroke join and cap methods
varying float v_length;


// The origin of fragment coordinates is located at bottom left .
// Hence, to convert it to top left, we need to know the height.
uniform float height;

void main()
{
    vec2 uv = gl_FragCoord.xy;
    float inside;

    vec3 p; // vector from fragment to vertex
    p = vec3(
            uv.x - v_pos.x,
            (height - uv.y) - v_pos.y,
            1.0
        );

    if(dot(v_tangentNext, p) < 0.0){
        // do nothing
    } else if(dot(v_tangentNext, p)/length(v_tangentNext) > v_length){
        // shift coordinates of p to next vertex
        p = vec3(
            uv.x - (v_pos.x + v_tangentNext.x),
            (height - uv.y) - (v_pos.y + v_tangentNext.y),
            1.0
        );
    } else{
        gl_FragColor = frag_color;
        return;
    }

    if(v_cap > 0.0){
        // render stroke cap
        if(v_cap_type == 0.0){ // PROJECT
            inside = 1.0;
        } else if(v_cap_type == 1.0){ // SQUARE
            inside = -1.0;
        }
        else if(v_cap_type == 2.0){ // ROUND
            inside = v_linewidth/2 - length(p);
        }
    } else{
        // render stroke join
        if(v_join_type == 0.0){ // MITER
            inside = 1.0;
        } else if(v_join_type == 1.0){ // BEVEL
            vec3 bisector = -normalize(normalize(v_tangentPrev) + normalize(v_tangentNext));
            inside = v_linewidth/2 - dot(p, bisector);

            if(inside > 0.0){
                inside = v_linewidth/2 - dot(p, -bisector);
            }
        }
        else if(v_join_type == 2.0){ // ROUND
            inside = v_linewidth/2 - length(p);
        }
    }

    if(inside > 0.0){
        gl_FragColor = frag_color;
    } else{
        // discard the fragment
        discard;
    }    
}
"""


src_default = ShaderSource(default_vertex_source, default_fragment_source)
src_texture = ShaderSource(texture_vertex_source, texture_fragment_source)
src_fbuffer = ShaderSource(fbuffer_vertex_source, fbuffer_fragment_source)
src_line = ShaderSource(stroke_vertex_source, stroke_fragment_source)



import unittest

from p5.pmath.curves import (bezier_point, bezier_tangent, curve_point,
        curve_tangent, quadratic_point)
from p5.pmath.vector import Point

class TestCurves(unittest.TestCase):
    def test_bezier_point(self):
        p1 = (30, 20)
        p2 = (80, 5)
        p3 = (80, 75)
        p4 = (30, 75)

        self.assertEqual(
            bezier_point(p1, p2, p3, p4, 0.5),
            Point(67.5, 41.875, 0))

        self.assertEqual(
            bezier_point(p1, p2, p3, p4, 0.9),
            Point(43.5, 73.055, 0))

    def test_bezier_tangent(self):
        p1 = (30, 20)
        p2 = (80, 5)
        p3 = (80, 75)
        p4 = (30, 75)

        self.assertEqual(
            bezier_tangent(p1, p2, p3, p4, 0.5),
            Point(0.0, 93.75, 0))

    def test_curve_point(self):
        p1 = (73, 24)
        p2 = (73, 61)
        p3 = (15, 65)
        p4 = (15, 65)
        self.assertEqual(curve_point(p1, p2, p3, p4, 0.5), Point(44.0, 65.3125, 0))

    def test_curve_tangent(self):
        p1 = (73, 24)
        p2 = (73, 61)
        p3 = (15, 65)
        p4 = (15, 65)
        self.assertEqual(curve_tangent(p1, p2, p3, p4, 0.5), Point(-72.5, 0.375, 0))

    def test_quadratic_point(self):
        p1 = (20, 20)
        p2 = (80, 20)
        p3 = (50, 50)
        self.assertEqual(quadratic_point(p1, p2, p3, 0.5), Point(57.5, 27.5, 0))

if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import xml.etree.ElementTree as etree
import math
import re
import numpy as np

from . import Color
from . import PShape
from . import primitives
from . import transforms
from ..pmath import matrix

default_values = { # default values of SVG attributes
	"stroke-width": 1,
	"stroke-color": Color(0, 0, 0), 
	"stroke-join": 0,
	"stroke-cap": 1,
	"stroke": "none",
	"fill": "none"
}

def get_style(element, style):
	# style parser for svg
	values = element.get("style")

	if element.get(style):
		return element.get(style)

	if values:
		for s in values.split(";"):
			if style in s:
				value = s.split(':')[1]
				
				if style == "stroke" or style == "fill":
					return value
				if style == "stroke-width":
					return int(value.replace("px", ""))
				if style == "stroke-opacity":
					return float(value)

	if style in default_values.keys():
		return default_values[style]
	else:
		return None

def parse_rect(element):
	width = float(element.get('width'))
	height = float(element.get('height'))
	x = float(element.get('x'))
	y = float(element.get('y'))

	fill = Color(get_style(element, "fill"))
	stroke_weight = get_style(element, "stroke-width")
	stroke = Color(get_style(element, "stroke"))
	stroke_cap = get_style(element, "stroke-cap")

	return PShape([
		(x, y),
		(x + width, y),
		(x + width, y + height),
		(x, y + height)
		], children=[], 
		fill_color=fill, stroke_weight=stroke_weight, 
		stroke_color=stroke, stroke_cap=stroke_cap,
		stroke_join=default_values["stroke-join"])

def parse_circle(element):
	cx = float(element.get('cx'))
	cy = float(element.get('cy'))
	r = float(element.get('r'))

	fill = Color(get_style(element, "fill"))
	stroke_weight = get_style(element, "stroke-width")
	stroke = Color(get_style(element, "stroke"))
	stroke_cap = get_style(element, "stroke-cap")

	return primitives.Arc(
		(cx, cy),
		(r/2, r/2),
		0, 2*math.pi,
		"CHORD",
		fill_color=fill, stroke_weight=stroke_weight, 
		stroke_color=stroke, stroke_cap=2)

def parse_line(element):
	x1 = float(element.get('x1'))
	y1 = float(element.get('y1'))
	x2 = float(element.get('x2'))
	y2 = float(element.get('y2'))

	fill = Color(get_style(element, "fill"))
	stroke_weight = get_style(element, "stroke-width")
	stroke = Color(get_style(element, "stroke"))
	stroke_cap = get_style(element, "stroke-cap")

	return PShape([(x1, y1), (x2, y2)], attribs='path', 
		fill_color=fill, stroke_weight=stroke_weight, 
		stroke_color=stroke, stroke_cap=stroke_cap,
		stroke_join=default_values["stroke-join"])

def parse_ellipse(element):
	cx = float(element.get('cx'))
	cy = float(element.get('cx'))
	rx = float(element.get('rx'))
	ry = float(element.get('ry'))

	fill = Color(get_style(element, "fill"))
	stroke_weight = get_style(element, "stroke-width")
	stroke = Color(get_style(element, "stroke"))
	stroke_cap = get_style(element, "stroke-cap")


	return primitives.Arc(
		(cx, cy),
		(rx/2, ry/2),
		0, 2*math.pi, 
		"CHORD", 
		fill_color=fill, stroke_weight=stroke_weight, 
		stroke_color=stroke, stroke_cap=2, stroke_join=default_values["stroke-join"])	

parser_function = {
	# tag: parser
	"rect": parse_rect,
	"circle": parse_circle,
	"line": parse_line,
	"ellipse": parse_ellipse,
	#"path": parse_path
}

def load_shape(filename):
	"""
	Loads the given .svg file and converts it into
	PShape object.

	:param filename: link to .svg file
    :type filename: str

	"""
	tree = etree.parse(filename)

	root = tree.getroot()
	if root.tag != "{http://www.w3.org/2000/svg}svg":
		raise TypeError('file %s does not seem to be a valid SVG file', filename)

	width = root.get('width')
	height = root.get('height')
	
	svg = transform_shape(parser(root))

	return svg

def transform_shape(element):
	# apply to current element
	element.apply_transform_matrix(element._transform_matrix)

	# apply to all its children
	for child in element.children:
		transform_shape(child)

	return element

def parser(element):
	shape = PShape([(0,0)], children=[])

	transform = element.get('transform')
	transform_matrix = np.identity(4)
	if transform:
		properties = re.findall(r"(\w+)\(([\w\.]+)[,\s]*([^(]*)\)", transform)
		for p in properties:
			if p[0] == "scale":
				mat = matrix.scale_transform(float(p[1]), float(p[2]))
				transform_matrix = transform_matrix.dot(mat)
			elif p[0] == "translate":
				mat = matrix.translation_matrix(float(p[1]), float(p[2]))
				transform_matrix = transform_matrix.dot(mat)

	shape.transform_matrix(transform_matrix)

	for e in element:
		tag = e.tag.replace('{http://www.w3.org/2000/svg}', "")
		if tag in parser_function.keys():
			shape.add_child(parser_function[tag](e))
		elif tag == "g":
			shape.add_child(parser(e))
		else:
			continue

	return shape

def shape(shape, x=0, y=0):
	"""
	Draws shapes to the display window

	:param shape: shape object
    :type shape: PShape

	:param x: x-coordinate of the shape
    :type x: float

    :param y: y-coordinate of the shape
    :type y: float
	"""
	with transforms.push_matrix():
		transforms.translate(x, y)
		primitives.draw_shape(shape)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from . import _p5

# we should be parsing sys args here and expose a series of commands
# similar to processing-java.
#
# relevant output from `processing-java --help`:
#
#     --help               Show this help text. Congratulations.
#    
#     --sketch=<name>      Specify the sketch folder (required)
#     
#     --output=<name>      Specify the output folder (optional and
#                          cannot be the same as the sketch folder.)
#                          
#     --force              The sketch will not build if the output
#                          folder already exists, because the contents
#                          will be replaced. This option erases the
#                          folder first. Use with extreme caution!
#                          
#     --build              Preprocess and compile a sketch into .class files.
#     
#     --run                Preprocess, compile, and run a sketch.
#     
#     --present            Preprocess, compile, and run a sketch in presentation mode.
#    
#     --export             Export an application.
#     
#     --no-java            Do not embed Java. Use at your own risk!
#     
#     --platform           Specify the platform (export to application only).
#                          Should be one of 'windows', 'macosx', or 'linux'.
#
                                                                                    



import unittest

import numpy as np
from p5.pmath.vector import Vector
from p5.pmath.utils import PI

a = Vector(2, 3, 4)
b = Vector(5, 6, 7)
c = Vector(1, 1)

class TestVector(unittest.TestCase):

    def test_properties(self):
        self.assertEqual(a.x, 2)
        self.assertEqual(a.y, 3)
        self.assertEqual(a.z, 4)

    def test_distance(self):
        self.assertTrue(np.allclose(a.distance(b), 5.196152))

    def test_lerp(self):
        self.assertEqual(a.lerp(b, 0.5), Vector(3.5, 4.5, 5.5))

    def test_arithmetic(self):
        self.assertEqual(a + b, Vector(7, 9, 11))
        self.assertEqual(a - b, Vector(-3, -3, -3))
        self.assertEqual(a*2, Vector(4, 6, 8))
        self.assertEqual(-a, Vector(-2, -3, -4))

    def test_dot(self):
        self.assertEqual(a.dot(b), 56)

    def test_cross(self):
        self.assertEqual(a.cross(b), Vector(-3, 6, -3))

    def test_angle(self):
        self.assertTrue(np.allclose(c.angle, PI/4))

    def test_angle_between(self):
        self.assertTrue(np.allclose(a.angle_between(b), 0.13047689))

if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from .structure import *
from .shape import *
from .primitives import *
from .primitives3d import *
from .transforms import *
from .color import *
from .attribs import *
from .image import *
from .font import *
from .constants import *
from .vertex import *
from .svg import *


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from contextlib import contextmanager


from . import primitives
from . import color
from . import p5

@contextmanager
def push_style():
    """Save the current style settings and then restores them on exit.

    The 'style' information consists of all the parameters controlled
    by the following functions (the ones indicated by an asterisks '*'
    aren't available yet):

    - background
    - fill, no_fill
    - stroke, no_stroke
    - rect_mode
    - ellipse_mode
    - shape_mode
    - color_mode
    - tint
    - (*) stroke_weight
    - (*) stroke_cap
    - (*) stroke_join
    - (*) image_mode
    - (*) text_align
    - (*) text_font
    - (*) text_mode
    - (*) text_size
    - (*) text_leading
    - (*) emissive
    - (*) specular
    - (*) shininess
    - (*) ambient

    """
    prev_background_color = p5.renderer.background_color
    prev_fill_color = p5.renderer.fill_color
    prev_fill_enabled = p5.renderer.fill_enabled
    prev_stroke_enabled = p5.renderer.stroke_enabled
    prev_stroke_color = p5.renderer.stroke_color
    prev_tint_color = p5.renderer.tint_color
    prev_tint_enabled = p5.renderer.tint_enabled

    prev_ellipse_mode = primitives._ellipse_mode
    prev_rect_mode = primitives._rect_mode
    prev_shape_mode = primitives._shape_mode

    prev_color_mode = color.color_parse_mode
    prev_color_range = color.color_range

    yield

    p5.renderer.background_color = prev_background_color
    p5.renderer.fill_color = prev_fill_color
    p5.renderer.fill_enabled = prev_fill_enabled
    p5.renderer.stroke_color = prev_stroke_color
    p5.renderer.stroke_enabled = prev_stroke_enabled
    p5.renderer.tint_color = prev_tint_color
    p5.renderer.tint_enabled = prev_tint_enabled

    primitives._ellipse_mode = prev_ellipse_mode
    primitives._rect_mode = prev_rect_mode
    primitives._shape_mode = prev_shape_mode

    color.prev_color_parse_mode = prev_color_mode
    color.prev_color_range = prev_color_range

def pop_style():
    push_style()



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Functions to work with Bezier curves and Catmul-Rom splines.

"""

from functools import wraps
import math

from ..pmath import Point

__all__ = [
    # BEZIER METHODS
    'bezier_point', 'bezier_tangent', 'bezier_detail',

    # CURVE METHODS
    'curve_point', 'curve_tangent', 'curve_detail', 'curve_tightness',

    # QUADRATIC METHODS
    'quadratic_point'
]

curve_resolution = 20
curve_tightness_amount = 0

bezier_resolution = 20

curve_basis_matrix = [
    [-0.5, 1.5, -1.5, 0.5],
    [1, -2.5, 2, -0.5],
    [-0.5, 0, 0.5, 0],
    [0, 1, 0, 0]
]

def typecast_arguments_as_points(func):
    """Typecast all but the last argument of the function as Points."""
    @wraps(func)
    def decorated(*args, **kwargs):
        new_args = [Point(*arg) for arg in args[:-1]]
        new_args.append(args[-1])
        ret_value = func(*new_args, **kwargs)
        return ret_value
    return decorated

def bezier_detail(detail_value):
    """Change the resolution used to draw bezier curves.

    :param detail_value: New resolution to be used.
    :type detail_value: int
    """
    global bezier_resolution
    bezier_resolution = max(1, detail_value)

@typecast_arguments_as_points
def bezier_point(start, control_1, control_2, stop, parameter):
    """Return the coordinate of a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: 3-tuple.

    :param control_1: The first control point of the bezier curve
    :type control_1: 3-tuple.

    :param control_2: The second control point of the bezier curve
    :type control_2: 3-tuple.

    :param stop: The end point of the bezier curve
    :type stop: 3-tuple.

    :param parameter: The parameter for the required location along
        the curve. Should be in the range [0.0, 1.0] where 0 indicates
        the start of the curve and 1 indicates the end of the curve.
    :type parameter: float

    :returns: The coordinate of the point along the bezier curve.
    :rtype: Point (namedtuple with x, y, z attributes)

    """
    t = parameter
    t_ = 1 - parameter

    P = [start, control_1, control_2, stop]
    coeffs = [t_*t_*t_, 3*t*t_*t_,  3*t*t*t_, t*t*t]

    x = sum(pt.x * c for pt, c in zip(P, coeffs))
    y = sum(pt.y * c for pt, c in zip(P, coeffs))

    return Point(x, y)

@typecast_arguments_as_points
def bezier_tangent(start, control_1, control_2, stop, parameter):
    """Return the tangent at a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: 3-tuple.

    :param control_1: The first control point of the bezier curve
    :type control_1: 3-tuple.

    :param control_2: The second control point of the bezier curve
    :type control_2: 3-tuple.

    :param stop: The end point of the bezier curve
    :type stop: 3-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the bezier
        curve.
    :rtype: Point (namedtuple with x, y, z attributes)

    """
    t = parameter
    tangent = lambda a, b, c, d: 3*t*t*(3*b - 3*c + d - a) + \
                                 6*t*(a - 2*b + c) + \
                                 3*(b - a)
    x = tangent(start.x, control_1.x, control_2.x, stop.x)
    y = tangent(start.y, control_1.y, control_2.y, stop.y)
    return Point(x, y)

def _reinit_curve_matrices():
    # TODO: Add basis matrices for faster tessellation.
    pass

def curve_detail(detail_value):
    """Change the resolution used to draw bezier curves.

    :param detail_value: New resolution to be used.
    :type detail_value: int
    """
    global curve_resolution
    curve_resolution = detail_value

def curve_tightness(amount):
    """Change the curve tightness used to draw curves.

    :param amount: new curve tightness amount.
    :type amount: int
    """
    global curve_tightness_amount
    curve_tightness_amount = amount
    _reinit_curve_matrices()

@typecast_arguments_as_points
def curve_point(point_1, point_2, point_3, point_4, parameter):
    """Return the coordinates of a point along a curve.

    :param point_1: The first control point of the curve.
    :type point_1: 3-tuple.

    :param point_2: The second control point of the curve.
    :type point_2: 3-tuple.

    :param point_3: The third control point of the curve.
    :type point_3: 3-tuple.

    :param point_4: The fourth control point of the curve.
    :type point_4: 3-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: Point (namedtuple with x, y, z attributes)

    """
    t = parameter
    basis = curve_basis_matrix
    P = [point_1, point_2, point_3, point_4]

    coeffs = [sum(t**(3 - i) * basis[i][j] for i in range(4)) for j in range(4)]

    x = sum(pt.x * c for pt, c in zip(P, coeffs))
    y = sum(pt.y * c for pt, c in zip(P, coeffs))

    return Point(x, y)

@typecast_arguments_as_points
def curve_tangent(point_1, point_2, point_3, point_4, parameter):
    """Return the tangent at a point along a curve.

    :param point_1: The first control point of the curve.
    :type point_1: 3-tuple.

    :param point_2: The second control point of the curve.
    :type point_2: 3-tuple.

    :param point_3: The third control point of the curve.
    :type point_3: 3-tuple.

    :param point_4: The fourth control point of the curve.
    :type point_4: 3-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the curve.
    :rtype: Point (namedtuple with x, y, z attributes)

    """
    t = parameter
    basis = curve_basis_matrix
    P = [point_1, point_2, point_3, point_4]

    coeffs = [
        sum((3 - i)*(t**(2 - i)) * basis[i][j] for i in range(3))
        for j in range(4)
    ]

    x = sum(pt.x * c for pt, c in zip(P, coeffs))
    y = sum(pt.y * c for pt, c in zip(P, coeffs))

    return Point(x, y)

@typecast_arguments_as_points
def quadratic_point(start, control, stop, parameter):
    """Return the coordinates of a point along a bezier curve.

    :param point_1: The start point of the curve.
    :type point_1: 3-tuple.

    :param point_3: The control point of the curve.
    :type point_3: 3-tuple.

    :param point_4: The end point of the curve.
    :type point_4: 3-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: Point (namedtuple with x, y, z attributes)

    """
    t = parameter
    t_ = 1 - parameter

    P = [start, control, stop]
    coeffs = [t_*t_, 2*t*t_, t*t]

    x = sum(pt.x * c for pt, c in zip(P, coeffs))
    y = sum(pt.y * c for pt, c in zip(P, coeffs))

    return Point(x, y)

# Set the default values.
bezier_detail(20)
curve_detail(20)
curve_tightness(0)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import builtins
from collections import namedtuple
from enum import IntEnum

Position = namedtuple('Position', ['x', 'y'])

handler_names = [ 'key_pressed', 'key_released', 'key_typed',
                  'mouse_clicked', 'mouse_double_clicked',
                  'mouse_dragged', 'mouse_moved',
                  'mouse_pressed', 'mouse_released', 'mouse_wheel',]


class VispyButton(IntEnum):
    LEFT = 1
    RIGHT = 2
    MIDDLE = 3


class MouseButton:
    """An abstraction over a set of mouse buttons.

    :param buttons: list of mouse buttons pressed at the same time.
    :type buttons: str list

    """
    def __init__(self, buttons):
        button_names = {
            VispyButton.LEFT: 'LEFT',
            VispyButton.RIGHT: 'RIGHT',
            VispyButton.MIDDLE: 'MIDDLE',
        }

        self._buttons = buttons
        self._button_names = [button_names[bt] for bt in self._buttons]

    @property
    def buttons(self):
        self._button_names

    def __eq__(self, other):
        button_map = {
            'CENTER': VispyButton.MIDDLE,
            'MIDDLE': VispyButton.MIDDLE,
            'LEFT':  VispyButton.LEFT,
            'RIGHT': VispyButton.RIGHT,
        }
        if isinstance(other, str):
            return button_map.get(other.upper(), -1) in self._buttons
        return self._buttons == other._buttons

    def __neq__(self, other):
        return not (self == other)

    def __repr__(self):
        fstr = ', '.join(self.buttons)
        return "MouseButton({})".format(fstr)

    __str__ = __repr__


class Key:
    """A higher level abstraction over a single key.

    :param name: The name of the key; ENTER, BACKSPACE, etc.
    :type name: str

    :param text: The text associated with the given key. This
        corresponds to the symbol that will be "typed" by the given
        key.
    :type name: str

    """
    def __init__(self, name, text=''):
        self.name = name.upper()
        self.text = text

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.name or other == self.text
        return self.name == other.name and self.text == other.text

    def __neq__(self, other):
        return not (self == other)

    def __str__(self):
        if self.text.isalnum():
            return self.text
        else:
            return self.name

    def __repr__(self):
        return "Key({})".format(self.name)


class Event:
    """A generic sketch event.

    :param modifers: The set of modifiers held down at the time of the
        event.
    :type modifiers: str list

    :param pressed: If the key/button is held down when the event
        occurs.
    :type pressed: bool

    """
    def __init__(self, raw_event, active=False):
        self._modifiers = list(map(lambda k: k.name, raw_event.modifiers))
        self._active = active
        self._raw = raw_event

    @property
    def modifiers(self):
        return self._modifiers

    @property
    def pressed(self):
        return self._active

    def is_shift_down(self):
        """Was shift held down during the event?

        :returns: True if the shift-key was held down.
        :rtype: bool

        """
        return 'Shift' in self._modifiers

    def is_ctrl_down(self):
        """Was ctrl (command on Mac) held down during the event?

        :returns: True if the ctrl-key was held down.
        :rtype: bool

        """
        return 'Control' in self._modifiers

    def is_alt_down(self):
        """Was alt held down during the event?

        :returns: True if the alt-key was held down.
        :rtype: bool

        """
        return 'Alt' in self._modifiers

    def is_meta_down(self):
        """Was the meta key (windows/option key) held down?

        :returns: True if the meta-key was held down.
        :rtype: bool

        """
        return  'Meta' in self._modifiers

    def _update_builtins(self):
        pass


class KeyEvent(Event):
    """Encapsulates information about a key event.

    :param key: The key associated with this event.
    :type key: str

    :param pressed: Specifies whether the key is held down or not.
    :type pressed: bool

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self._raw.key is not None:
            self.key = Key(self._raw.key.name, self._raw.text)
        else:
            self.key = Key('UNKNOWN')

    def _update_builtins(self):
        builtins.key_is_pressed = self.pressed
        builtins.key = self.key if self.pressed else None


class MouseEvent(Event):
    """A class that encapsulates information about a mouse event.

    :param x: The x-position of the mouse in the window at the time of
        the event.
    :type x: int

    :param y: The y-position of the mouse in the window at the time of
        the event.
    :type y: int

    :param position: Position of the mouse in the window at the time
        of the event.
    :type position: (int, int)

    :param change: the change in the x and y directions (defaults to
        (0, 0))
    :type change: (int, int)

    :param scroll: the scroll amount in the x and y directions
         (defaults to (0, 0)).
    :type scroll: (int, int)

    :param count: amount by which the mouse whell was dragged.
    :type count: int

    :param button: Button information at the time of the event.
    :type button: MouseButton

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        x, y = self._raw.pos
        x = max(min(builtins.width, x), 0)
        y = max(min(builtins.height, builtins.height - y), 0)
        dx, dy = self._raw.delta

        if (self._raw.press_event != None) and (self._raw.last_event != None):
            px, py = self._raw.press_event.pos
            cx, cy = self._raw.last_event.pos
            self.change = Position(cx - px, cy - py)
        else:
            self.change = Position(0, 0)

        self.x = max(min(builtins.width, x), 0)
        self.y = max(min(builtins.height, builtins.height - y), 0)

        self.position = Position(x, y)
        self.scroll = Position(int(dx), int(dy))

        self.count = self.scroll.y
        self.button = MouseButton(self._raw.buttons)

    def _update_builtins(self):
        builtins.pmouse_x = builtins.mouse_x
        builtins.pmouse_y = builtins.mouse_y
        builtins.mouse_x = self.x
        builtins.mouse_y = self.y
        builtins.mouse_is_pressed = self._active
        builtins.mouse_is_dragging = (self.change == (0, 0))
        builtins.mouse_button = self.button if self.pressed else None

    def __repr__(self):
        press = 'pressed' if self.pressed else 'not-pressed'
        return "MouseEvent({} at {})".format(press, self.position)

    __str__ = __repr__



#!/usr/bin/env python

import os
import sys

from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))

packages = ['p5']

requires = [
    'vispy',
    'glfw',
    'numpy',
    'Pillow',
    'triangle',
]

meta_data = {}
with open(os.path.join(here, 'p5', '__version__.py'), 'r', encoding='utf-8') as f:
    exec(f.read(), meta_data)

with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name=meta_data['__title__'],
    version=meta_data['__version__'],
    description=meta_data['__description__'],
    long_description=readme,
    long_description_content_type="text/x-rst",
    author=meta_data['__author__'],
    url=meta_data['__url__'],
    author_email=meta_data['__author_email__'],
    license=meta_data['__license__'],
    packages=find_packages(),
    package_data={'': ['LICENSE'], },
    package_dir={'p5': 'p5'},
    include_package_data=True,

    setup_requires=['numpy'],
    install_requires=requires,

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',

        'Topic :: Artistic Software',
        'Topic :: Education',

        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import functools
import math

from . import p5
from .geometry import Geometry

from ..pmath import matrix
from .primitives import draw_shape, _draw_on_return

# We use these in ellipse tessellation. The algorithm is similar to
# the one used in Processing and the we compute the number of
# subdivisions per ellipse using the following formula:
#
#    min(M, max(N, (2 * pi * size / F)))
#
# Where,
#
# - size :: is the measure of the dimensions of the circle when
#   projected in screen coordiantes.
#
# - F :: sets the minimum number of subdivisions. A smaller `F` would
#   produce more detailed circles (== POINT_ACCURACY_FACTOR)
#
# - N :: Minimum point accuracy (== MIN_POINT_ACCURACY)
#
# - M :: Maximum point accuracy (== MAX_POINT_ACCURACY)
#
MIN_POINT_ACCURACY = 20
MAX_POINT_ACCURACY = 200
POINT_ACCURACY_FACTOR = 10

@_draw_on_return
def box(width, height, depth, detail_x=1, detail_y=1):
    """
    Draw a plane with given a width and height

    :param width: width of the box
    :type width: float

    :param height: height of the box
    :type height: float
  
    :param depth: depth of the box
    :type depth: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 1
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 1
    :type detail_y: integer
    """

    geom = Geometry(detail_x, detail_y)

    cube_indices = [
        [0, 4, 2, 6], # -1, 0, 0],// -x
        [1, 3, 5, 7], # +1, 0, 0],// +x
        [0, 1, 4, 5], # 0, -1, 0],// -y
        [2, 6, 3, 7], # 0, +1, 0],// +y
        [0, 2, 1, 3], # 0, 0, -1],// -z
        [4, 5, 6, 7]  # 0, 0, +1] // +z
    ]

    geom.stroke_indices = [
        [0, 1],
        [1, 3],
        [3, 2],
        [6, 7],
        [8, 9],
        [9, 11],
        [14, 15],
        [16, 17],
        [17, 19],
        [18, 19],
        [20, 21],
        [22, 23]
    ]

    for i in range(len(cube_indices)):
        cube_index = cube_indices[i]
        v = i * 4
        for j in range(4):
            d = cube_index[j]

            octant = [
                ((d & 1) * 2 - 1) / 2,
                ((d & 2) - 1) / 2,
                ((d & 4) / 2 - 1) / 2
            ]

            geom.vertices.append(octant)
            geom.uvs.extend([j & 1, (j & 2) / 2])

        geom.faces.append([v, v + 1, v + 2])
        geom.faces.append([v + 2, v + 1, v + 3])

    geom.compute_normals()
    geom.make_triangle_edges()
    geom.matrix = matrix.scale_transform(width, height, depth)

    return geom

@_draw_on_return
def plane(width, height, detail_x=1, detail_y=1):
    """
    Draw a plane with given a width and height

    :param width: width of the plane
    :type width: float

    :param height: height of the plane
    :type height: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 1
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 1
    :type detail_y: integer
    """
    geom = Geometry(detail_x, detail_y)

    for i in range(detail_y + 1):
        v = i/detail_y
        for j in range(detail_x + 1):
            u = j/detail_x
            p = [u - 0.5, v - 0.5, 0]
            geom.vertices.append(p)
            geom.uvs.extend([u, v])

    geom.compute_faces()
    geom.compute_normals()
    geom.make_triangle_edges()
    geom.edges_to_vertices()
    geom.matrix = matrix.scale_transform(width, height, 1)

    return geom

def sphere(radius=50, detail_x=24, detail_y=16):
    """
    Draw a sphere with given radius

    :param radius: radius of circle
    :type radius: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 24
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 16
    :type detail_y: integer
    """

    return ellipsoid(radius, radius, radius, detail_x, detail_y)

@_draw_on_return
def ellipsoid(radius_x, radius_y, radius_z, detail_x=24, detail_y=24):
    """
    Draw an ellipsoid with given radius

    :param radius_x: x-radius of ellipsoid
    :type radius_x: float
    
    :param radius_y: y-radius of ellipsoid
    :type radius_y: float

    :param radius_z: z-radius of ellipsoid
    :type radius_z: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 24
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 16
    :type detail_y: integer
    """
    geom = Geometry(detail_x, detail_y)

    for i in range(detail_y + 1):
        v = i / detail_y
        phi = math.pi * v - math.pi/2
        cosPhi = math.cos(phi)
        sinPhi = math.sin(phi)

        for j in range(detail_x + 1):
            u = j / detail_x
            theta = 2 * math.pi * u
            cosTheta = math.cos(theta)
            sinTheta = math.sin(theta)
            p = [cosPhi * sinTheta, sinPhi, cosPhi * cosTheta]

            geom.vertices.append(p)
            geom.vertex_normals.append(p)
            geom.uvs.extend([u, v])

    geom.compute_faces()
    geom.make_triangle_edges()
    geom.edges_to_vertices()
    geom.matrix = matrix.scale_transform(radius_x, radius_y, radius_z)

    return geom

def truncated_cone(bottom_radius, top_radius, height, detail_x, detail_y, bottom_cap, top_cap):
    geom = Geometry(detail_x, detail_y)

    bottom_radius = 1 if bottom_radius <= 0 else bottom_radius
    top_radius = 0 if top_radius < 0 else top_radius
    height = bottom_radius if height <= 0 else height
    detail_x = 3 if detail_x < 3 else detail_x
    detail_y = 1 if detail_y < 1 else detail_y

    start = -2 if bottom_cap else 0
    end = detail_y + (2 if top_cap else 0)
    
    slant = math.atan2(bottom_radius - top_radius, height)
    sin_slant = math.sin(slant)
    cos_slant = math.cos(slant)

    for yy in range(start, end + 1):
        v = yy / detail_y
        y = height * v 
        ring_radius = 0 

        if yy < 0:
            # for the bottomCap edge
            y = 0
            v = 0
            ring_radius = bottom_radius
        elif yy > detail_y:
            # for the topCap edge
            y = height
            v = 1
            ring_radius = top_radius
        else:
            # for the middle
            ring_radius = bottom_radius + (top_radius - bottom_radius) * v

        if yy == -2 or yy == detail_y + 2:
            # center of bottom or top caps
            ring_radius = 0

        y -= height/2 # shift coordiate origin to the center of object
        for ii in range(detail_x):
            u = ii / detail_x
            ur = 2*math.pi*u
            sur = math.sin(ur)
            cur = math.cos(ur)

            geom.vertices.append([sur * ring_radius, y, cur * ring_radius])

            if yy < 0:
                vertex_normals = [0, -1, 0]
            elif yy > detail_y and top_radius:
                vertex_normals = [0, 1, 0]
            else:
                vertex_normals = [sur * cos_slant, sin_slant, cur * cos_slant]

            geom.vertex_normals.append(vertex_normals)
            geom.uvs.extend([u, v])

    start_index = 0
    if bottom_cap:
        for jj in range(detail_x):
            nextjj = (jj + 1) % detail_x
            geom.faces.append([
                start_index + jj,
                start_index + detail_x + nextjj,
                start_index + detail_x + jj
                ])

        start_index += detail_x*2

    for yy in range(detail_y):
        for ii in range(detail_x):
            nextii = (ii + 1) % detail_x
            geom.faces.append([
                start_index + ii,
                start_index + nextii,
                start_index + detail_x + nextii
                ])
            geom.faces.append([
                start_index + ii,
                start_index + detail_x + nextii,
                start_index + detail_x + ii
                ])

        start_index += detail_x

    if top_cap:
        start_index += detail_x
        for ii in range(detail_x):
            geom.faces.append([
                start_index + ii,
                start_index + (ii + 1) % detail_x,
                start_index + detail_x
                ])

    return geom

@_draw_on_return
def cylinder(radius=50, height=50, detail_x=24, detail_y=1, top_cap=True, bottom_cap=True):
    """
    Draw a cylinder with given radius and height

    :param radius: radius of the surface
    :type radius: float
    
    :param height: height of the cylinder
    :type height: float

    :param detail_x: Number of segments, the more segments the smoother geometry. Default is 24
    :type detail_x: integer

    :param detail_y: number of segments in y-dimension, the more segments the smoother geometry. Default is 1
    :type detail_y: integer

    :param bottom_cap: whether to draw the bottom of the cylinder
    :type bottom_cap: boolean

    :param top_cap: whether to draw the top of the cylinder
    :type top_cap: boolean
    """

    geom = truncated_cone(1, 1, 1, detail_x, detail_y, bottom_cap, top_cap)
    geom.matrix = matrix.scale_transform(radius, height, radius)

    geom.make_triangle_edges()
    geom.edges_to_vertices()

    return geom

@_draw_on_return
def cone(radius=50, height=50, detail_x=24, detail_y=1, cap=True):
    """
    Draw a cone with given radius and height

    :param radius: radius of the bottom surface
    :type radius: float
    
    :param height: height of the cone
    :type height: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 24
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 1
    :type detail_y: integer
    """
    geom = truncated_cone(1, 0, 1, detail_x, detail_y, cap, False)

    geom.make_triangle_edges()
    geom.edges_to_vertices()

    geom.matrix = matrix.scale_transform(radius, height, radius)
    return geom

@_draw_on_return
def torus(radius=50, tube_radius=10, detail_x=24, detail_y=16):
    """
    Draws torus on the window

    :param radius: radius of the whole ring
    :type radius: float
    
    :param tube_radius: radius of the tube
    :type tube_radius: float

    :param detail_x: Optional number of triangle subdivisions in x-dimension. Default is 24
    :type detail_x: integer

    :param detail_y: Optional number of triangle subdivisions in y-dimension. Default is 16
    :type detail_y: integer
    """
    tube_ratio = tube_radius / radius
    geom = Geometry(detail_x, detail_y)

    for i in range(detail_y + 1):
        v = i / detail_y
        phi = 2*math.pi*v 
        cosPhi = math.cos(phi)
        sinPhi = math.sin(phi)
        r = 1 + tube_ratio * cosPhi

        for j in range(detail_x + 1):
            u = j / detail_x
            theta = 2 * math.pi * u 
            cosTheta = math.cos(theta)
            sinTheta = math.sin(theta)

            geom.vertices.append([
                r * cosTheta,
                r * sinTheta, 
                tube_ratio * sinPhi
                ])

        n = [cosPhi * cosTheta, cosPhi * sinTheta, sinPhi]
        geom.vertex_normals.append(n)
        geom.uvs.extend([u, v])

    geom.compute_faces()
    geom.make_triangle_edges()
    geom.edges_to_vertices()
    geom.matrix = matrix.scale_transform(radius, radius, radius)

    return geom


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
3D geometry class for p5py
"""

import numpy as np
import math

class Geometry:
	"""
	Geometry class for all 3D shapes

    :param detail_x: number of triangle subdivisions in x-dimension
    :type detail_x: integer

    :param detail_y: number of triangle subdivisions in y-dimension
    :type detail_y: integer

	"""
	def __init__(self, detail_x=1, detail_y=1):
		self.vertices = []

		self.line_vertices = []

		self.line_normals = []

		self.vertex_normals = []

		self.faces = []

		self.uvs = []
		# a 2D array containing edge connectivity pattern for create line vertices
		# based on faces for most objects
		self.edges = []
		self.detail_x = detail_x
		self.detail_y = detail_y

		self.stroke_indices = []

		self.matrix = np.identity(4)

	def reset(self):
		"""
		Reset geometry parameters
		"""
		self.vertices = []
		self.line_vertices = []
		self.line_normals = []
		self.vertex_normals = []
		self.faces = []
		self.uvs = []
		self.edges = []

	def compute_faces(self):
		"""
		Adds the faces for the geometry for predefined order of vertices
		"""
		self.faces = []
		sliceCount = self.detail_x + 1
		for i in range(self.detail_y):
			for j in range(self.detail_x):
				a = i * sliceCount + j
				b = i * sliceCount + j + 1
				c = (i + 1) * sliceCount + j + 1
				d = (i + 1) * sliceCount + j
				self.faces.append([a, b, d])
				self.faces.append([d, b, c])

	def make_triangle_edges(self):
		"""
		Adds the edges to the geometry based on the faces
		"""
		self.edges = []
		for j in range(len(self.faces)):
			self.edges.append([self.faces[j][0], self.faces[j][1]])
			self.edges.append([self.faces[j][1], self.faces[j][2]])
			self.edges.append([self.faces[j][2], self.faces[j][0]])

	def get_face_normal(self, faceId):
		"""
		Returns the normal for a given face
		"""
		face = self.faces[faceId]
		vA = np.array(self.vertices[face[0]])
		vB = np.array(self.vertices[face[1]])
		vC = np.array(self.vertices[face[2]])
		ab = vB - vA
		ac = vC - vA
		n = np.cross(ab, ac)
		ln = np.linalg.norm(n)

		sinAlpha = ln / (np.linalg.norm(ab) * np.linalg.norm(ac))
		if sinAlpha > 1: 
			sinAlpha = 1

		return n*math.sin(sinAlpha)/ln

	def compute_normals(self):
		"""
		Compute normals for every vertex
		"""
		self.vertex_normals = []
		for iv in range(len(self.vertices)):
			self.vertex_normals.append([0, 0, 0])

		for f in range(len(self.faces)):
			face = self.faces[f]
			face_normal = self.get_face_normal(f)

			for fv in range(3):
				vertex_index = face[fv]
				self.vertex_normals[vertex_index] += face_normal

		for iv in range(len(self.vertices)):
			self.vertex_normals[iv] = self.vertex_normals[iv]/np.linalg.norm(self.vertex_normals[iv])

	def edges_to_vertices(self):
		self.line_vertices = []
		self.line_normals = []

		for i in range(len(self.edges)):
			begin = self.vertices[self.edges[i][0]]
			end = self.vertices[self.edges[i][1]]

			direction = np.array(end) - np.array(begin)
			direction = direction/np.linalg.norm(direction)
			direction = direction.tolist()

			a = begin
			b = begin
			c = end
			d = end
			dirAdd = direction
			dirSub = direction
			dirAdd.append(1)
			dirSub.append(-1)

			self.line_normals.extend([dirAdd, dirSub, dirAdd, dirAdd, dirSub, dirSub])
			self.line_vertices.extend([a, b, c, c, b, d])


#
# part of p5: a python package based on processing
# copyright (c) 2017-2019 abhik pal
#
# this program is free software: you can redistribute it and/or modify
# it under the terms of the gnu general public license as published by
# the free software foundation, either version 3 of the license, or
# (at your option) any later version.
#
# this program is distributed in the hope that it will be useful, but
# without any warranty; without even the implied warranty of
# merchantability or fitness for a particular purpose. see the gnu
# general public license for more details.
#
# you should have received a copy of the gnu general public license
# along with this program. if not, see <http://www.gnu.org/licenses/>.
#
"""Userspace functions"""

import __main__

import math
import numpy as np
import builtins
from functools import wraps

from vispy import app

from .base import Sketch
from .events import handler_names

from ..core import p5
from ..pmath import matrix

from ..core.constants import *
from .renderer2d import Renderer2D
from .renderer3d import Renderer3D

__all__ = ['no_loop', 'loop', 'redraw', 'size', 'title', 'no_cursor',
           'cursor', 'exit', 'draw', 'setup', 'run', 'save_frame', 'save']

builtins.width = 360
builtins.height = 360
builtins.pixel_x_density = 1
builtins.pixel_y_density = 1

builtins.title = "p5"
builtins.frame_count = -1
builtins.frame_rate = None
builtins.focused = True

builtins.mouse_button = None
builtins.mouse_is_pressed = False
builtins.mouse_is_dragging = False
builtins.mouse_x = 0
builtins.mouse_y = 0
builtins.pmouse_x = 0
builtins.pmouse_y = 0
builtins.key = None
builtins.key_is_pressed = False

builtins.pixels = None


def _fix_interface(func):
    """Make sure that `func` takes at least one argument as input.

    :returns: a new function that accepts arguments.
    :rtype: func
    """
    @wraps(func)
    def fixed_func(*args, **kwargs):
        return_value = func()
        return return_value

    if func.__code__.co_argcount == 0:
        return fixed_func
    else:
        return func

def draw():
    """Continuously execute code defined inside.

    The `draw()` function is called directly after `setup()` and all
    code inside is continuously executed until the program is stopped
    (using `exit()`) or `no_loop()` is called.

    """
    pass

def setup():
    """Called to setup initial sketch options.

    The `setup()` function is run once when the program starts and is
    used to define initial environment options for the sketch.

    """
    pass

def run(sketch_setup=None, sketch_draw=None, frame_rate=60, mode="P2D"):
    """Run a sketch.

    if no `sketch_setup` and `sketch_draw` are specified, p5 automatically
    "finds" the user-defined setup and draw functions.

    :param sketch_setup: The setup function of the sketch (None by
         default.)
    :type sketch_setup: function

    :param sketch_draw: The draw function of the sketch (None by
        default.)
    :type sketch_draw: function

    :param frame_rate: The target frame rate for the sketch.
    :type frame_rate: int :math:`\geq 1`

    """

    # get the user-defined setup(), draw(), and handler functions.
    if sketch_setup is not None:
        setup_method = sketch_setup
    elif hasattr(__main__, 'setup'):
        setup_method = __main__.setup
    else:
        setup_method = setup

    if sketch_draw is not None:
        draw_method = sketch_draw
    elif hasattr(__main__, 'draw'):
        draw_method = __main__.draw
    else:
        draw_method = draw

    handlers = dict()
    for handler in handler_names:
        if hasattr(__main__, handler):
            hfunc = getattr(__main__, handler)
            handlers[handler] = _fix_interface(hfunc)

    if mode == "P2D":
        p5.renderer = Renderer2D()
        p5.mode = 'P2D'
    elif mode == "P3D":
        p5.mode = 'P3D'
        p5.renderer = Renderer3D()
    else:
        raise ValueError("Invalid Mode %s" % mode)

    p5.sketch = Sketch(setup_method, draw_method, handlers, frame_rate)

    physical_width, physical_height = p5.sketch.physical_size
    width, height = p5.sketch.size

    builtins.pixel_x_density = physical_width // width
    builtins.pixel_y_density = physical_height // height

    p5.sketch.timer.start()

    app.run()
    exit()

def title(new_title):
    """Set the title of the p5 window.

    :param new_title: new title of the window.
    :type new_title: str

    """
    builtins.title = new_title
    p5.sketch.title = new_title

def size(width, height):
    """Resize the sketch window.

    :param width: width of the sketch window.
    :type width: int

    :param height: height of the sketch window.
    :type height: int

    """
    builtins.width = int(width)
    builtins.height = int(height)
    p5.sketch.size = (builtins.width, builtins.height)

    # update the look at matrix coordinates according to sketch size
    if p5.mode == "P3D":
        p5.renderer.lookat_matrix = matrix.look_at(
        np.array([0, 0, height/math.tan(math.pi/6)]), 
        np.array([0, 0, 0]), 
        np.array([0, 1, 0]))

def no_loop():
    """Stop draw() from being continuously called.

    By default, the sketch continuously calls `draw()` as long as it
    runs. Calling `no_loop()` stops draw() from being called the next
    time. Note that this only prevents execution of the code inside
    `draw()` and the user can manipulate the screen contents through
    event handlers like `mouse_pressed()`, etc.

    """
    p5.sketch.looping = False
    p5.sketch.redraw = True

def loop():
    """Make sure `draw()` is being called continuously.

    `loop()` reverts the effects of `no_loop()` and allows `draw()` to
    be called continously again.

    """
    p5.sketch.looping = True

def redraw():
    """Call `draw()` once.

    If looping has been disabled using `no_loop()`, `redraw()` will
    make sure that `draw()` is called *exactly* once.

    """
    if not p5.sketch.looping:
        p5.sketch.redraw = True

def exit(*args, **kwargs):
    """Exit the sketch.

    `exit()` overrides Python's builtin exit() function and makes sure
    that necessary cleanup steps are performed before exiting the
    sketch.

    :param args: positional argumets to pass to Python's builtin
        `exit()` function.

    :param kwargs: keyword-arguments to pass to Python's builtin
        `exit()` function.
    """
    if not (p5.sketch is None):
        p5.sketch.show(visible=False)
        app.quit()
    p5.exit(*args, **kwargs)

def no_cursor():
    """Hide the mouse cursor.
    """
    # window.set_mouse_visible(False)
    raise NotImplementedError

def cursor(cursor_type='ARROW'):
    """Set the cursor to the specified type.

    :param cursor_type: The cursor type to be used (defaults to
        'ARROW'). Should be one of: {'ARROW','CROSS','HAND', 'MOVE',
        'TEXT', 'WAIT'}
    :type cursor_type: str

    """
    # cursor_map = {
    #     'ARROW': window.CURSOR_DEFAULT,
    #     'CROSS': window.CURSOR_CROSSHAIR,
    #     'HAND': window.CURSOR_HAND,
    #     'MOVE': window.CURSOR_SIZE,
    #     'TEXT': window.CURSOR_TEXT,
    #     'WAIT': window.CURSOR_WAIT
    # }
    # selected_cursor = cursor_map.get(cursor_type, 'ARROW')
    # cursor = window.get_system_mouse_cursor(selected_cursor)
    # window.set_mouse_visible(True)
    # window.set_mouse_cursor(cursor)
    raise NotImplementedError

def save(filename='screen.png'):
    """Save an image from the display window. 

    Saves an image from the display window. Append a file extension to
    the name of the file, to indicate the file format to be used.If no
    extension is included in the filename, the image will save in PNG
    format and .png will be added to the name. By default, these files
    are saved to the folder where the sketch is saved. Alternatively,
    the files can be saved to any location on the computer by using an
    absolute path (something that starts with / on Unix and Linux, or
    a drive letter on Windows).

    :param filename: Filename of the image (defaults to
        ``screen.png``)

    :type filename: str

    """
    # TODO: images saved using ``save()`` should *not* be numbered.
    # --abhikpal (2018-08-14)
    p5.sketch.screenshot(filename)

def save_frame(filename="screen.png"):
    """Save a numbered sequence of images whenever the function is run.

    Saves a numbered sequence of images, one image each time the
    function is run. To save an image that is identical to the display
    window, run the function at the end of :meth:`p5.draw` or within
    mouse and key events such as :meth:`p5.mouse_pressed` and
    :meth:`p5.key_pressed`. 

    If save_frame() is used without parameters, it will save files as
    screen-0000.png, screen-0001.png, and so on. Append a file
    extension, to indicate the file format to be used. Image files are
    saved to the sketch's folder. Alternatively, the files can be
    saved to any location on the computer by using an absolute path
    (something that starts with / on Unix and Linux, or a drive letter
    on Windows).

    :param filename: name (or name with path) of the image file.
        (defaults to ``screen.png``)

    :type filename: str

    """
    # todo: allow setting the frame number in the file name of the
    # saved image (instead of using the default sequencing) --abhikpal
    # (2018-08-14)
    p5.sketch.queue_screenshot(filename)
